﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using System.Drawing;

namespace oprava_obrazku
{
    public class RepairImage
    {
        private Bitmap bmp;
        private bool[,] map, done;
        private Graphics bmp_gr;

        public Bitmap Image{ get { return bmp; } }

        public RepairImage(Bitmap bmp)
        {
            this.bmp = bmp;
            map = new bool[bmp.Width, bmp.Height];
            done = new bool[bmp.Width, bmp.Height];
            bmp_gr = Graphics.FromImage(bmp);
        }

        public void Repair() //oprava obrázku
        {            
            for (int y = 0; y < bmp.Height; y++)
            {
                for (int x = 0; x < bmp.Width; x++)
                {
                    //je pozice poškozená?
                    if (compare_color(bmp.GetPixel(x, y)))
                    {
                        //priorita - počet barevných bodů kolem aktuálního bodu
                        int counts = get_counts(x, y);

                        //pokud je priorita vetší 2
                        if (counts > 2)
                        {
                            //obarvení pozice
                            Color new_color = get_color(x, y); //získání barvy z okolních bodů
                            bmp_gr.FillRectangle(new SolidBrush(new_color), x, y, 1, 1);
                            done[x, y] = true; //hotový bod
                        }
                        else
                            set_pos(new Point(x, y)); //procházení plochy pomocí vlny
                    }
                }
            }
        }

        private int get_counts(int ox, int oy) //vrací počet barevných bodů kolem pozice - pro zkrácení tzv. priorita
        {
            int count = 0;

            for (int i = 0; i < 4; i++)
            {
                int x = ox, y = oy;

                if (i == 0) //up
                    y--;
                else if (i == 1) //left
                    x--;
                else if (i == 2) //down
                    y++;
                else //right
                    x++;

                if ((x < 0 || x >= bmp.Width || y < 0 || y >= bmp.Height) || (get_color_from_pixel(x, y) != Color.FromArgb(0, 0, 0, 0))) count++;
            }

            return count;
        }

        private void set_pos(Point loc) //projití větší plochy pomocí vlny
        {
            int x = loc.X;
            int y = loc.Y;

            List<Point> positions = new List<Point>();
            positions.Add(new Point(x, y));
        
            //nastavení priority bodům
            for (int was_count = 0; was_count < positions.Count; was_count++)
            {
                Point actual = positions[was_count]; //aktuální pozice

                x = actual.X;
                y = actual.Y;

                map[x, y] = true; //na teto pzici jsem uz byl

                for (int i = 0; i < 4; i++)
                {
                    int new_x = x, new_y = y;

                    if (i == 0) //up
                        new_y--;
                    else if (i == 1) //left
                        new_x--;
                    else if (i == 2) //down
                        new_y++;
                    else //right
                        new_x++;

                    //pokud není nová pozice mimo obrázek
                    if (new_x >= 0 && new_x < bmp.Width && new_y >= 0 && new_y < bmp.Height)
                    {
                        //pokud jsme na pozici už nebyly a je pokud je pozice poškozená
                        if (!map[new_x, new_y] && compare_color(bmp.GetPixel(new_x, new_y)))
                        {
                            map[new_x, new_y] = true; //na teto pzici jsem uz byl
                            positions.Add(new Point(new_x, new_y)); //pridání nové pozice
                        }
                    }
                }
            }
       
            int count = 4;
            bool change = false;
            x = 0; y = 0;

            //obarvování bodů
            while (count > 0)
            { 
                change = false;

                //hledání všech pozic s aktuálně vyhledávanou prioritou
                for (int i = 0; i < positions.Count; i++)
                {
                    Point pos = positions[i];
                    if (get_counts(pos.X, pos.Y) == count) //pokud se priorita schoduje 
                    {
                        x = pos.X; y = pos.Y;

                        //obarvení
                        Color new_color = get_color(x, y);
                        bmp_gr.FillRectangle(new SolidBrush(new_color), x, y, 1, 1);

                        positions.RemoveAt(i); //odstranění pozice ze seznamu

                        done[x, y] = true; //hotová pozice
                        change = true; //provedla se změna
                    }
                }

                //pokud se provedla nějaká změna
                if (!change) //ne
                    count--; //hledat pozice s menší prioritou
                else //ano 
                    count = 4; //hledat pozice se stejnou prioritou 
            }
        }

        private bool compare_color(Color c1) //je barva poškozená?
        { 
            //má hodnotu alfa 0
            if (c1.A == 0) //ano
            {
                if (c1.R == 0) // ...
                {
                    if (c1.G == 0)
                    {
                        if (c1.B == 0)
                            return true;
                        else
                            return false;
                    }
                    else
                        return false;
                }
                else
                    return false;
            }
            else //ne
                return false;
        }

        private Color get_color(int x, int y) //smíchání barvy z okolních pozic
        {
            //získání barev ze všech směrů
            Color up = get_color_from_pixel(x, y - 1);
            Color left = get_color_from_pixel(x - 1, y);
            Color down = get_color_from_pixel(x, y + 1);
            Color right = get_color_from_pixel(x + 1, y);

            //míchání barev
            Color c1 = blend(up, left); 
            Color c2 = blend(down, right);

            return blend(c1, c2);
        }

        private Color blend(Color c1, Color c2) //smíchání 2 barev
        {
            //pokud jedna z barev je poškozená - nemíchat barvy
            if (compare_color(c1)) 
                return c2;
            else if (compare_color(c2))
                return c1;
            else //ne
            {
                //vrací průměr RGB složek každé barvy
                return Color.FromArgb(255,
                    (int)((c1.R + c2.R) / 2),
                    (int)((c1.G + c2.G) / 2),
                    (int)((c1.B + c2.B) / 2)
                );
            }
        }

        private Color get_color_from_pixel(int x, int y) //získání barvy z pozice
        {
            //pokud je pozice mimo obrázek ... vrací poškozenou barvu
            if (x < 0 || x >= bmp.Width || y < 0 || y >= bmp.Height)
                return Color.FromArgb(0, 0, 0, 0);
            else
                return bmp.GetPixel(x, y);
        }
    }
}
